import math
import cv2
import numpy as np
import sys


class gResMCSegPre:
    """
          Formal class using Deep Residual Multiscale Segmenter for prediction

         # Examples

         ```python
             #

             dkwargs={'inc_step': 2, 'up2deconv': True, 'activation': 'relu', 'nfeature': 0, 'dropout': 0.5, 'batchnorm': True,
                'cropsize': None, 'ismaxpool': True, 'isconvcon': False, 'k_initializer': 'glorot_uniform',
                'residual': False, 'residualAutoType': 0, 'multsctype': None, 'mctotallevels': 10,
                'mcbothcoder': False, 'mcfilters': [16, 16,16,16,16,16,16,16,16,16,16,16,16,16],'cropsize':16}
             start_filter=32
             modelCls = gResMCSeg(input_shape=(256, 256,20), start_filter= start_filter, depth=depth, **dkwargs)
             model = modelCls.ResMCNet()
             model.summary()
             model.compile(optimizer="adam", loss= K_nJIBCE_loss,
                   metrics=['accuracy',K_nJI, K_nJIBCE_loss])
        ```
          # Arguments
             patchsize: Input size of a patch sample ;
             bordersize: Border size to filter out border effect  ;
             overprop: Overlapping proportion for two neighbor images, default: 0, no overlapping ;

          #Output:
             Predicted mask matrix (np array type)
             Predicted mask matrix (np array type)
        """
    def __init__(self,patchsize,bordersize,overprop=0):
        self.patchsize=patchsize
        self.border_size=bordersize
        self.overprop=overprop

    def preAImgBi(self,image, model):

        """
          Function to predict binary mask based on the trained model
          :param image: input image ;
          :param model: the trained model ;
          :return: the matrix of predicted mask (0,1)
        """
        imarray = np.array(image)
        imarray = cv2.copyMakeBorder(imarray, self.border_size, self.border_size, self.border_size,
                                     self.border_size, cv2.BORDER_REFLECT)
        nrow, ncol, _ = imarray.shape
        if (nrow < ncol and nrow < self.patchsize) or (nrow > ncol and ncol < self.patchsize):
            return None
        imgres = np.zeros((image.shape[0], image.shape[1], 2))
        # img=imarray/255.0
        stride = math.floor((1 - self.overprop) * self.patchsize)
        for ir in range(0, nrow + 1, stride):
            for jc in range(0, ncol + 1, stride):
                ifrm = ir
                jfrm = jc
                it = ifrm + self.patchsize + self.border_size * 2
                jt = jfrm + self.patchsize + self.border_size * 2
                if it > nrow:
                    it = nrow
                    ifrm = nrow - self.patchsize - self.border_size * 2
                if jt > ncol:
                    jt = ncol
                    jfrm = ncol - self.patchsize - self.border_size * 2
                #                    it = it + self.border_size * 2
                #                    jt = jt + self.border_size * 2
                imgB = imarray[ifrm:it, jfrm:jt, :]
                imgB = imgB[None, :, :, :]
                # print(ir,jc,jfrm,jt,imgB.shape)
                res = model.predict(imgB)
                i2c = (ifrm, ifrm + self.patchsize)
                j2c = (jfrm, jfrm + self.patchsize)
                ti2 = (self.border_size, self.patchsize + self.border_size)
                tj2 = (self.border_size, self.patchsize + self.border_size)
                #        print(np.max(res), np.min(res))
                imgres[i2c[0]:i2c[1], j2c[0]:j2c[1], 0] = imgres[i2c[0]:i2c[1], j2c[0]:j2c[1], 0] \
                                                          + res[0, ti2[0]:ti2[1], tj2[0]:tj2[1], 0]
                imgres[i2c[0]:i2c[1], j2c[0]:j2c[1], 1] = imgres[i2c[0]:i2c[1], j2c[0]:j2c[1], 1] + 1.0
        imgres = imgres[:, :, 0] / imgres[:, :, 1]
        print("total res:", np.max(imgres), np.min(imgres))
        return imgres


    def preAImgMulti(self, image,model,ncls=9):
        """
          Function to predict multiclass mask based on the trained model
            :param image: input image ;
            :param model: the trained model ;
            :param ncls: number of classes ;
            :return: the matrix of predicted mask, ncls dimensions
        """
        imarray = np.array(image)
        imarray = cv2.copyMakeBorder(imarray, self.border_size, self.border_size, self.border_size,
                                 self.border_size, cv2.BORDER_REFLECT)
        nrow, ncol, _ = imarray.shape
        if (nrow < ncol and nrow < self.patchsize) or (nrow > ncol and ncol < self.patchsize):
            return
        imgres = np.zeros((image.shape[0], image.shape[1], ncls + 1))
        stride = math.floor((1 - self.overprop) * self.patchsize)
        for ir in range(0, nrow + 1, stride):
            for jc in range(0, ncol + 1, stride):
                ifrm = ir
                jfrm = jc
                it = ifrm + self.patchsize + self.border_size * 2
                jt = jfrm + self.patchsize + self.border_size * 2
                if it > nrow:
                    it = nrow
                    ifrm = nrow - self.patchsize - self.border_size * 2
                if jt > ncol:
                    jt = ncol
                    jfrm = ncol - self.patchsize - self.border_size * 2
                imgB = imarray[ifrm:it, jfrm:jt, :]
                imgB = imgB[None, :, :, :]
                # print(ir,jc,jfrm,jt,imgB.shape)
                res = model.predict(imgB)
                i2c = (ifrm, ifrm + self.patchsize)
                j2c = (jfrm, jfrm + self.patchsize)
                imgres[i2c[0]:i2c[1], j2c[0]:j2c[1], 0:ncls] = imgres[i2c[0]:i2c[1], j2c[0]:j2c[1], 0:ncls] + res
                imgres[i2c[0]:i2c[1], j2c[0]:j2c[1], ncls] = imgres[i2c[0]:i2c[1], j2c[0]:j2c[1], ncls] + 1.0
        imgres = imgres[:, :, 0:ncls] / imgres[:, :, ncls][:, :, None]
        print("total res:", np.max(imgres), np.min(imgres))
        pindex = np.argmax(imgres, axis=2)
        imgres = np.zeros(imgres.shape)
        for c in range(imgres.shape[0]):
            for r in range(imgres.shape[1]):
                imgres[c, r, pindex[c, r]] = 1
        return imgres





from keras.utils import conv_utils
from keras.backend.common import normalize_data_format
from keras.engine import InputSpec
from keras.engine import Layer
import keras
import tensorflow as tf

class ResizeLayer(Layer):
    """ResizeLayer class: Resize Images to a specified size

    # Arguments
        output_size: Size of output layer width and height
        data_format: A string,
               one of `channels_last` (default) or `channels_first`.
              The ordering of the dimensions in the inputs.
               `channels_last` corresponds to inputs with shape
               `(batch, height, width, channels)` while `channels_first`
                  corresponds to inputs with shape
                `(batch, channels, height, width)`.
                 It defaults to the `image_data_format` value found in your
                  Keras config file at `~/.keras/keras.json`.
                  If you never set it, then it will be "channels_last".

    # Input shape
        - If `data_format='channels_last'`:
            4D tensor with shape:
            `(batch_size, rows, cols, channels)`
        - If `data_format='channels_first'`:
            4D tensor with shape:
            `(batch_size, channels, rows, cols)`

    # Output shape
        - If `data_format='channels_last'`:
            4D tensor with shape:
            `(batch_size, pooled_rows, pooled_cols, channels)`
        - If `data_format='channels_first'`:
            4D tensor with shape:
            `(batch_size, channels, pooled_rows, pooled_cols)`
    """
    def __init__(self, output_dim=(1, 1), data_format=None, **kwargs):
        super(ResizeLayer, self).__init__(**kwargs)
        data_format = normalize_data_format(data_format)
        self.output_dim = conv_utils.normalize_tuple(output_dim, 2, 'output_dim')
        self.data_format = normalize_data_format(data_format)
        self.input_spec = InputSpec(ndim=4)
        self.padding='same'


    def build(self, input_shape):
        self.input_spec = [InputSpec(shape=input_shape)]

    def compute_output_shape(self, input_shape):
        if self.data_format == 'channels_first':
            return (input_shape[0], input_shape[1], self.output_dim[0], self.output_dim[1])
        elif self.data_format == 'channels_last':
            return (input_shape[0], self.output_dim[0], self.output_dim[1], input_shape[3])

    def _resize_fun(self, inputs, data_format):
        try:
            assert keras.backend.backend() == 'tensorflow'
            assert self.data_format == 'channels_last'
        except AssertionError:
            print("Only tensorflow backend is supported for the resize layer and accordingly 'channels_last' ordering")
        output = tf.image.resize_images(inputs, self.output_dim)
        return output

    def call(self, inputs):
        output = self._resize_fun(inputs=inputs, data_format=self.data_format)
        return output

    def get_config(self):
        config = {'output_dim': self.output_dim,
           #       'padding': self.padding,
                  'data_format': self.data_format}
        base_config = super(ResizeLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


import keras.backend as K
from keras.backend import binary_crossentropy
import numpy as np

color_dict={ 0: (255, 255, 255), # white Background
             1: (0, 0, 0), # Black; Roads
             2: (0, 255, 0),# Green : Grass
             3: (255, 255, 0), # Yellow : Rails
             4: (150, 80, 0), # dark brown : Bare soil
             5: (0, 125, 0), #dark green: trees
             6: (0, 0, 150),# blue water
             7: (100, 100, 100), # gray buildings
             8: (150, 150, 255)} # blue-purple: pools

color_dict_cv2={ 0: (255, 255, 255), # white Background
             1: (0, 0, 0), # Black; Roads
             2: (0, 255, 0),# Green : Grass
             3: (0, 255, 255), # Yellow : Rails
             4: (0, 80, 150), # dark brown : Bare soil
             5: (0, 125, 0), #dark green: trees
             6: (150, 0, 0),# blue water
             7: (100, 100, 100), # gray buildings
             8: (255, 150, 150)} # blue-purple: pools

def rgb_to_onehot(rgb_arr, color_dict):
    num_classes = len(color_dict)
    shape = rgb_arr.shape[:2]+(num_classes,)
    arr = np.zeros( shape, dtype=np.int8 )
    for i, cls in enumerate(color_dict):
        arr[:,:,i] = np.all(rgb_arr.reshape( (-1,3) ) == color_dict[i], axis=1).reshape(shape[:2])
    return arr

def onehot_to_rgb(onehot, color_dict):
    single_layer = np.argmax(onehot, axis=-1)
    output = np.zeros( onehot.shape[:2]+(3,) )
    for k in color_dict.keys():
        output[single_layer==k] = color_dict[k]
    return np.uint8(output)

MAXPIXELVAL=2026.0

def dice_loss(y_true, y_pred):
    smooth = 1.
    y_true_f = K.flatten(y_true)
    y_pred_f = K.flatten(y_pred)
    intersection = y_true_f * y_pred_f
    score = (2. * K.sum(intersection) + smooth) / (K.sum(y_true_f) + K.sum(y_pred_f) + smooth)
    return 1. - score


def bce_dice_loss(y_true, y_pred):
    return binary_crossentropy(y_true, y_pred) + dice_loss(y_true, y_pred)


def jaccard_coef1(y_true, y_pred):
    smooth = 1
    intersection = K.sum(y_true * y_pred, axis=[0, -1, -2])
    sum_ = K.sum(y_true + y_pred, axis=[0, -1, -2])
    jac = (intersection + smooth) / (sum_ - intersection + smooth)
    return K.mean(jac)


def jaccard_coef_int(y_true, y_pred):
    smooth = 1
    y_pred_pos = K.round(K.clip(y_pred, 0, 1))
    intersection = K.sum(y_true * y_pred_pos, axis=[0, -1, -2])
    sum_ = K.sum(y_true + y_pred_pos, axis=[0, -1, -2])
    jac = (intersection + smooth) / (sum_ - intersection + smooth)
    return K.mean(jac)

def jaccard_coef(y_true, y_pred):
    smooth = 1
    y_true_f = K.flatten(y_true)
    y_pred_f = K.flatten(y_pred)
    intersection = K.sum(y_true_f * y_pred_f)
    return (intersection + smooth) / (K.sum(y_true_f) + K.sum(y_pred_f) - intersection + smooth)




import numpy as np
from sklearn.metrics import confusion_matrix

import keras.backend as K
from keras.backend import binary_crossentropy
import tensorflow as tf
from resmcseg._metrics import *

def k_nji(y_true, y_pred):
    """
    Function of normal Jaccard Index (keras)
       :param y_true: ground truth mask ;
       :param y_pred: predicted mask  ;
       :return: value of normal JI
    """
    smooth = 1
    y_true_f = K.flatten(y_true)
    y_pred_f = K.flatten(y_pred)
    intersection = K.sum(y_true_f * y_pred_f)
    return (intersection + smooth) / (K.sum(y_true_f) + K.sum(y_pred_f) - intersection + smooth)

def k_nji_loss(y_true, y_pred):
    """
    Function of log loss based on normal Jaccard Index (keras)
       :param y_true: ground truth mask ;
       :param y_pred: predicted mask  ;
       :return: loss fucntion value
    """
    return -K.log(k_nji(y_true, y_pred))

def k_njibce_loss(y_true, y_pred):
    """
    Function of log loss based on normal Jaccard Index and binary entropy  (keras)
      :param y_true: ground truth mask ;
      :param y_pred: predicted mask  ;
      :return: loss fucntion value
    """
    return -K.log(k_nji(y_true, y_pred)) + binary_crossentropy(y_pred, y_true)


def k_ji(y_true, y_pred):
    """
    Function of -non-normal Jaccard Index  (keras)
      :param y_true: ground truth mask ;
      :param y_pred: predicted mask  ;
      :return: JI value
    """
    smooth = 1
    y_pred_pos = K.round(K.clip(y_pred, 0, 1))
    intersection = K.sum(y_true * y_pred_pos, axis=[0, -1, -2])
    sum_ = K.sum(y_true + y_pred_pos, axis=[0, -1, -2])
    jac = (intersection + smooth) / (sum_ - intersection + smooth)
    return K.mean(jac)

def k_dice_loss(y_true, y_pred):
    """
    Function of dice loss  (keras)
      :param y_true: ground truth mask ;
      :param y_pred: predicted mask  ;
      :return: loss fucntion value
    """
    smooth = 1.
    y_true_f = K.flatten(y_true)
    y_pred_f = K.flatten(y_pred)
    intersection = y_true_f * y_pred_f
    score = (2. * K.sum(intersection) + smooth) / (K.sum(y_true_f) + K.sum(y_pred_f) + smooth)
    return 1. - score

def k_bce_logdice_loss(y_true, y_pred):
    """
    Function of dice and binary corss entropy loss  (keras)
      :param y_true: ground truth mask ;
      :param y_pred: predicted mask  ;
      :return: loss fucntion value
    """
    return binary_crossentropy(y_true, y_pred) - K.log(1. - K_dice_loss(y_true, y_pred))

def compute_iou(y_pred, y_true,labels=[0,1]):
    """
    mean IOU calculated based on ground truth and predicted values  (np)
      :param y_true: ground truth mask ;
      :param y_pred: predicted mask  ;
      :return: mean IOU
    """
    # ytrue, ypred is a flatten vector
    y_pred = y_pred.flatten()
    y_true = y_true.flatten()
    current = confusion_matrix(y_true, y_pred, labels=labels)
    # compute mean iou
    intersection = np.diag(current)
    ground_truth_set = current.sum(axis=1)
    predicted_set = current.sum(axis=0)
    union = ground_truth_set + predicted_set - intersection
    IoU = intersection / union.astype(np.float32)
    return np.mean(IoU)

def compute_iou_conf(current):
    """
    IOU calculated based on ground truth and predicted values  (np)
      :param y_true: ground truth mask ;
      :param y_pred: predicted mask  ;
      :return: mean IOU and each class's IOU
    """
    intersection = np.diag(current)
    ground_truth_set = current.sum(axis=1)
    predicted_set = current.sum(axis=0)
    union = ground_truth_set + predicted_set - intersection
    IoU = intersection / union.astype(np.float32)
    return (np.mean(IoU),IoU)

def jaccard(x, y):
    x = np.asarray(x, np.bool)  # Not necessary, if you keep your data
    y = np.asarray(y, np.bool)  # in a boolean array already!
    return np.double(np.bitwise_and(x, y).sum()) / np.double(np.bitwise_or(x, y).sum())


def mean_iou(y_true, y_pred):
    prec = []
    for t in np.arange(0.5, 1.0, 0.05):
        y_pred_ = tf.to_int32(y_pred > t)
        score, up_opt = tf.metrics.mean_iou(y_true, y_pred_, 2)
        K.get_session().run(tf.local_variables_initializer())
        with tf.control_dependencies([up_opt]):
            score = tf.identity(score)
        prec.append(score)
    return K.mean(K.stack(prec), axis=0)

def mean_iouC(y_true, y_pred):
    prec = []
    for t in np.arange(0.5, 1.0, 0.05):
        y_pred_ = tf.to_int32(y_pred > t)
        score, up_opt = tf.metrics.mean_iou(y_true, y_pred_, 9)
        K.get_session().run(tf.local_variables_initializer())
        with tf.control_dependencies([up_opt]):
            score = tf.identity(score)
        prec.append(score)
    return K.mean(K.stack(prec), axis=0)

def iou(y_true, y_pred, label: int):
    """
    Return the Intersection over Union (IoU) for a given label.
    Args:
    y_true: the expected y values as a one-hot
    y_pred: the predicted y values as a one-hot or softmax output
    label: the label to return the IoU for
    Returns:
    the IoU for the given label
    """
    # extract the label values using the argmax operator then
    # calculate equality of the predictions and truths to the label
    y_true = K.cast(K.equal(K.argmax(y_true), label), K.floatx())
    y_pred = K.cast(K.equal(K.argmax(y_pred), label), K.floatx())
    # calculate the |intersection| (AND) of the labels
    intersection = K.sum(y_true * y_pred)
    # calculate the |union| (OR) of the labels
    union = K.sum(y_true) + K.sum(y_pred) - intersection
    # avoid divide by zero - if the union is zero, return 1
    # otherwise, return the intersection over union
    return K.switch(K.equal(union, 0), 1.0, intersection / union)

def miou(y_true, y_pred):
    """
    Return the Intersection over Union (IoU) score.
    Args:
    y_true: the expected y values as a one-hot
    y_pred: the predicted y values as a one-hot or softmax output
    Returns:
    the scalar IoU value (mean over all labels)
    """
    # get number of labels to calculate IoU for
    num_labels = K.int_shape(y_pred)[-1] - 1
    # initialize a variable to store total IoU in
    mean_iou = K.variable(0)
    # iterate over labels to calculate IoU for
    for label in range(num_labels):
        mean_iou = mean_iou + iou(y_true, y_pred, label)
    # divide total IoU by number of labels to get mean IoU
    return mean_iou / num_labels



from resmcseg.util.googledloader import GoogleDriveDownloader as gdd

def downloadPretrainedModel(dataset,type=None,destination=None):
   """
    Function to download the pretrained models from Google cloud platform
      :param dataset: only trained models can be used by two dataset, i.e., DSTL and Zurich;
      :param type: model for the type of segmentation, useful for binary semantic segmentation of DSTL,
              Buildings, Trees,Vehicles, Crops and Roads;
              Models trained for the following data preprocessing:
                  DSTL: bands order: RGB, M, A, P (all bands used), A normalized by dividing by 16384,
                         the others normalized by dividing by 2048;
                  ZURICH: band orders: NIR-R-G-B, normalzied by dividing by 2026
      :param destinationFile: the destination file to be saved, default will be in the tmp path!
      :return: None
    """
   urlMapping={'DSTL':{'Trees':'1nWcHxlrSbpt6NhtOyRuoDxuKYL6Wi26d','Buildings':'1SjjlYPqGc1dICWb2o2wk40N5fgumdbns',
                       'Crops':'1ULlFc5mD8G9b6o1tdy8OjlNZiU0FZG5m','Roads':'1s1O8N5rnPrm7aupTEX2oZkFrTCm3dptg',
                       'Vehicles':'1ptvGriYgU49AJMFiu8CPcS51edsFRcbG'}, 'ZURICH':'1JXfnkfNNxHWsvYbGYQg_NmhP4hJggAW7'}
   if dataset not in urlMapping.keys():
       print("Only pretrained models for DSTL and ZURICH available!!!")
       return
   file_id=None
   if dataset=='ZURICH':
       if destination is None:
           destination='/tmp/ZURICH_model.h5'
       file_id=urlMapping[dataset]
   elif dataset=='DSTL':
       if type is None or type not in urlMapping[dataset].keys():
           print("Only accepting the types of Buildings, Trees, Vehicles, Crops and Roads for DSTL e!!!")
           return
       else:
           if destination is None:
               destination='/tmp/DSLT_'+type+'_model.h5'
           file_id=urlMapping[dataset][type]
   else:
       return
   if file_id is not None:
       gdd.download_file_from_google_drive(file_id=file_id,dest_path=destination,unzip=False)



# -*- coding: utf-8 -*-
""" Library of Autoencoder-based Residual Deep Network

The python library of Deep Residual Multiscale Segmenter (autonet).
Current version just supports the KERAS package of deep learning and
will extend to the others in the future.

Major modules
        model:gResMCSeg: major class to obtain a deep extensive residual multiscale
                  FCN. You can setup its aruments. See the class and its
                  member functions' help for details.
              gResMCSegPre: major class to make semantic segmentation for binary and multi class.
              pretrainedmodel: function to download the pretrained models using the DSTL and
                            ZURICH datasets from the Google cloud
        util: segmetrics: main metrics including jaccard index, MIoU, and loss functions etc.
              helper: helper functions including color mapping etc.
        data  function to access one image for Zurich to test the model's prediction.

Github source: https://github.com/lspatial/
Author: Lianfa Li
Date: 2018-10-01

"""

#import pkgutil
#__path__ = pkgutil.extend_path(__path__, __name__)

from resmcseg import model
from resmcseg import data
from resmcseg import util


__version__ = '0.1.62'




import pkg_resources
import tifffile as tiff
from osgeo import gdal
import numpy as np

from resmcseg.util.helper import MAXPIXELVAL,rgb_to_onehot,color_dict

def aimgload(name='zurich'):
    """
    Function to obtain the sample data.
    :param name: string, the name for each of two datasets.
          'zurich': One image from the Zurich dataset
          'dstl': not available now due to too large data
    :return: img and mask
    """
    if name!='zurich':
        print("Not supported for other data")
        return (None,None)
    fl=pkg_resources.resource_filename(__name__, '/zh17.tif')
    img = tiff.imread(fl)
    img = img.astype(np.float32) / MAXPIXELVAL
    fl = pkg_resources.resource_filename(__name__, '/zh17_gt.tif')
    gt = gdal.Open(fl).ReadAsArray()
    gt = np.rollaxis(gt, 0, 3)
    gt_onehot = rgb_to_onehot(gt, color_dict)
    print(gt_onehot.shape)
    return (img, gt_onehot)



# -*- coding: utf-8 -*-
"""
Major class: gAutoUnetClass

This class provides the functionality of deep residual multiscale
segmenter for remotely sensed images based on the KERAS environment.
The users need to set up their arguments of the model for construction
of the deep FCN. These parameters include number of layers, number of
nodes for each decoding layer, type of residuals, initializer, dropout
 rate, flag for batch normalization, regularizer, and multiscale type
 to obtain a residual network.

Author: Lianfa Li
Date: 2019-07-29

"""


from keras.models import Model
from keras.layers import Input, Lambda, concatenate, Concatenate,Conv2D, MaxPooling2D,RepeatVector,Reshape,add
from keras.layers import UpSampling2D, BatchNormalization, Dropout, Cropping2D, ZeroPadding2D,Conv2DTranspose
import math
from keras.layers.advanced_activations import ELU,ReLU
from .resizelayer import ResizeLayer
import keras.backend  as K
from keras.utils.training_utils import multi_gpu_model
import tensorflow as tf

class gResMCSeg:
    """
      Formal class of Deep Residual Multiscale Segmenter
      This class provides flexible architecture for deep extensive residual multiscale segmenter

     # Examples

     ```python
         # Set the dict type for multiple key-value paramaters :

         dkwargs={'inc_step': 2, 'up2deconv': True, 'activation': 'relu', 'nfeature': 0, 'dropout': 0.5, 'batchnorm': True,
            'cropsize': None, 'ismaxpool': True, 'isconvcon': False, 'k_initializer': 'glorot_uniform',
            'residual': False, 'residualAutoType': 0, 'multsctype': None, 'mctotallevels': 10,
            'mcbothcoder': False, 'mcfilters': [16, 16,16,16,16,16,16,16,16,16,16,16,16,16],'cropsize':16}
         start_filter=32
         modelCls = gResMCSeg(input_shape=(256, 256,20), start_filter= start_filter, depth=depth, **dkwargs)
         model = modelCls.ResMCNet()
         model.summary()
         model.compile(optimizer="adam", loss= K_nJIBCE_loss,
               metrics=['accuracy',K_nJI, K_nJIBCE_loss])
    ```
      # Arguments
         input_shape: shape of the input images including width, height and channel etc.;
         start_filter: number of filters for the starting layer ;
         depth: Number of the layers ;
         kwargs: other key-values parameters including
              'inc_step': incremental step for the number of filters for each decoding layer, default: 2,
               'up2deconv': use of upsampling rather than deconvolutions, default: True,
               'activation': activation function for the hidden layer, default: 'relu',
               'nfeature': number of features for each image sample, default: 0,
               'dropout': dropout rate, default: 0.5,
               'batchnorm': use of batch normalization, default: True,
               'cropsize': use of crop size for the input image, default: None,
               'ismaxpool': use of maximum pooling, default: True,
               'isconvcon': use of concatenation rather than residual connection for shortcut, default:  False,
               'k_initializer': initializer for the parameters, default: 'glorot_uniform',
               'residual': use of the residual unit, default: True,
               'residualAuto': use of nested residual connections, default: True,
               'multsctype': type of multiscale, four options:{resize, dilated (ASPP), both, None}, default:None,
               'mctotallevels': number of multiscle layers, default: 10,
               'mcbothcoder': False,
               'mcfilters': number of multiscale filter, default: None, setup by filters of each convolutions
      #Output:
         The model with the input and output tensor
    """

    def __init__(self,input_shape,start_filter,depth,**kwargs):
        self.input_shape=input_shape
        self.start_filter=start_filter
        self.depth=depth
        self.inc_step=kwargs['inc_step'] if 'inc_step' in  kwargs else 2
        self.up2deconv=kwargs['up2deconv'] if 'up2deconv' in  kwargs else True
        self.activation=kwargs['activation'] if 'activation' in  kwargs else 'relu'
        self.nfeature=kwargs['nfeature'] if 'nfeature' in  kwargs else 0
        self.dropout=kwargs['dropout'] if 'dropout' in  kwargs else 0.5
        self.batchnorm=kwargs['batchnorm'] if 'batchnorm' in  kwargs else True
        self.cropsize=kwargs['cropsize'] if 'cropsize' in  kwargs else None
        self.ismaxpool=kwargs['ismaxpool'] if 'ismaxpool' in  kwargs else True
        self.isconvcon=kwargs['isconvcon'] if 'isconvcon' in  kwargs else False
        self.k_initializer= kwargs['k_initializer'] if 'k_initializer' in  kwargs else 'glorot_uniform'
        self.residual=kwargs['residual'] if 'residual' in  kwargs else False
        self.residualAutoType=kwargs['residualAutoType'] if 'residualAutoType' in  kwargs else 0
        self.multsctype=kwargs['multsctype'] if 'multsctype' in  kwargs else None
        self.mctotallevels=kwargs['mctotallevels'] if 'mctotallevels' in  kwargs else 10
        self.mcbothcoder=kwargs['mcbothcoder'] if 'mcbothcoder' in  kwargs else False
        self.mcfilters=kwargs['mcfilters'] if 'mcfilters' in  kwargs else None

        self.nmaxpool =-1
        self.mclevel = 1


    def conv2dBlock(self, inlayer, nfilter, dropout=0):
        """
          Function to obtain one decoding or encoding building unit (consisting of multiple convolutions)
          :param inlayer: input layer ;
          :param nfilter: number of filters ;
          :param dropout: droput rate, default: 0 (no dropout layer applied)
          :return: the output layer of the encoding or decoding layer
        """
        if self.residual:
            #shortcut = Conv2D(nfilter, 3, activation=self.activation, padding='same',
            #                      kernel_initializer=self.k_initializer)(inlayer)
            shortcut=inlayer
            #shortcut = BatchNormalization()(shortcut) if self.batchnorm else shortcut
        layer = Conv2D(nfilter, 3, activation=self.activation, padding='same',kernel_initializer=self.k_initializer)(shortcut if self.residual else inlayer)
        layer = BatchNormalization()(layer) if self.batchnorm else layer
        if self.residual:
            shortcut=layer
        layer = Dropout(dropout)(layer) if dropout else layer
        layer = Conv2D(nfilter, 3, activation=self.activation, padding='same',kernel_initializer=self.k_initializer)(layer)
        layer = BatchNormalization()(layer) if self.batchnorm else layer
        if self.isconvcon :
             return Concatenate()([inlayer, layer])
 #            return add([inlayer, layer])
        if self.residual:
            layer=add([shortcut,layer])
            #layer=ReLU()(layer) if self.activation == 'relu' else ELU()(layer)
            return layer
        return layer

    def levelsBlock(self, inlayer, nfilter, depth, dropout,input_fea_layer=None):
        """
          Function to recursively construct the autoencoder architecture (encoding-coding-decoding)
          :param inlayer: input layer ;
          :param nfilter: number of filters ;
          :param depth: number of decoding units ;
          :param dropout: droput rate, default: 0 (no dropout layer applied)
          :param input_fea_layer: input of a feature, default: None  ;
          :return: the output tensor of this autoencoder
        """
        if depth > 0:
            layer = self.conv2dBlock(inlayer, nfilter)
            if self.ismaxpool:
                mlayer = MaxPooling2D()(layer)
                self.nmaxpool+=1
                if self.multsctype is not None:
                    mlayer=self.embeddedMcLayer(mlayer)
            else:
                mlayer = Conv2D(nfilter, 3, strides=2, padding='same',kernel_initializer=self.k_initializer)(layer)
            mlayer = self.levelsBlock(mlayer, int(self.inc_step * nfilter),depth - 1, dropout,input_fea_layer)
            if self.up2deconv:
                mlayer = UpSampling2D()(mlayer)
                if self.mcbothcoder and self.multsctype is not None:
                    mlayer = self.embeddedMcLayer(mlayer)
                mlayer = Conv2D(nfilter, 2, activation=self.activation, padding='same',kernel_initializer=self.k_initializer)(mlayer)
            else:
                mlayer = Conv2DTranspose(nfilter, 3, strides=2, activation=self.activation,
                                         padding='same',kernel_initializer=self.k_initializer)(mlayer)
            if self.residualAutoType==1:
                clayer = add([layer, mlayer])
            elif self.residualAutoType==0:
                clayer = Concatenate()([layer, mlayer])
            elif self.residualAutoType==2:
                clayer = add([layer, mlayer])
                clayer = Concatenate()([clayer, mlayer])
            elif self.residualAutoType==3:
                clayer = add([layer, mlayer])
                clayer = Concatenate()([clayer, layer])
            mlayer = self.conv2dBlock(clayer, nfilter)
        else:
            if self.nfeature==0 or self.nmaxpool<1:
                mlayer = self.conv2dBlock(inlayer, nfilter, dropout)
            else:
                krow = int(self.input_shape[0]/math.pow(2, self.nmaxpool))
                kcol = int(self.input_shape[1]/math.pow(2, self.nmaxpool))
                f_repeat = RepeatVector(krow *kcol)(input_fea_layer)
                f_conv = Reshape((krow, kcol, self.nfeature))(f_repeat)
                flayer = concatenate([inlayer, f_conv], -1)
                mlayer = self.conv2dBlock(flayer, nfilter, dropout)
        return mlayer

    def embeddedMcLayer(self,inlayer):
        """
          Function to multiscale input
          :param inlayer: input layer ;
          :return: this multiscale layer's output to be linked to its corresponding decoding layer
        """
        tshp = K.int_shape(inlayer)
        nfilter= int(tshp[3] / 4)
        if self.mcfilters is not None:
            nfilter=self.mcfilters[self.mclevel-1]
        if self.mclevel <= self.mctotallevels:
            if self.multsctype=='resize':
                mclayer = self.MultiScalePreLayer(nfilter=nfilter, output_dim=(tshp[1], tshp[2]))
                mlayer = Concatenate()([inlayer, mclayer])
                self.mclevel = self.mclevel + 1
            elif self.multsctype=='dilated':
                mclayer = self.MultiScaleDilateLayer(nfilter=nfilter, output_dim=(tshp[1], tshp[2]))
                mlayer = Concatenate()([inlayer, mclayer])
                self.mclevel = self.mclevel + 1
            elif self.multsctype=='both':
                mclayer = self.MultiScalePreLayer(nfilter=nfilter, output_dim=(tshp[1], tshp[2]))
                dilatedlayer = self.MultiScaleDilateLayer(nfilter=nfilter, output_dim=(tshp[1], tshp[2]))
                mlayer = Concatenate()([inlayer, dilatedlayer,mclayer])
                self.mclevel = self.mclevel + 1
            else:
                mlayer=inlayer
        return mlayer

    def MultiScalePreLayer(self, nfilter, output_dim):
        """
        Function to resizing multiscale layer
        :param nfilter: number of filters ;
         :param output_dim: dimension of the output ;
        :return: this multiscale layer's output
        """
        rzlayer = ResizeLayer(output_dim=output_dim)(self.inputlayer)
        outlayer = self.conv2dBlock(rzlayer, nfilter)
        return outlayer

    def MultiScaleDilateLayer(self, nfilter, output_dim):
        """
        Function to ASPP multiscale layer
        :param nfilter: number of filters ;
         :param output_dim: dimension of the output ;
        :return: this multiscale layer's output
        """
        layer = self.inputlayer
        dilations = [1, 4, 8, 16]
        aspp1 = Conv2D(nfilter, 3, activation='relu', padding='same', kernel_initializer=self.k_initializer,
                       dilation_rate=dilations[0])(layer)
        aspp1 = BatchNormalization()(aspp1)
        aspp2 = Conv2D(nfilter, 3, activation='relu', padding='same', kernel_initializer=self.k_initializer,
                       dilation_rate=dilations[1])(layer)
        aspp2 = BatchNormalization()(aspp2)
        aspp3 = Conv2D(nfilter, 3, activation='relu', padding='same', kernel_initializer=self.k_initializer,
                       dilation_rate=dilations[2])(layer)
        aspp3 = BatchNormalization()(aspp3)
        aspp4 = Conv2D(nfilter, 3, activation='relu', padding='same', kernel_initializer=self.k_initializer,
                       dilation_rate=dilations[3])(self.inputlayer)
        aspp4 = BatchNormalization()(aspp4)
        outlayer = Concatenate()([aspp1, aspp2, aspp3,aspp4])
        #outlayer = add([aspp1, aspp2, aspp3, aspp4])
        outlayer = BatchNormalization()(outlayer)
        outlayer = ReLU()(outlayer)
        outlayer = ResizeLayer(output_dim=output_dim)(outlayer)
        return outlayer


    def cropLayer(self,inlayer):
        """
        Function to cropping an input layer
        :param inlayer: input layer;
        :return: this output layer
        """
        if self.cropsize != None:
            croplayer = Cropping2D(cropping=((self.cropsize,self.cropsize),(self.cropsize,self.cropsize)))(inlayer)
            clayer = BatchNormalization(axis=3)(croplayer) if self.batchnorm else croplayer
            outlayer = ReLU()(clayer) if self.activation == 'relu' else ELU()(clayer)
            return outlayer
        return inlayer

    def ResMCNet(self,outfilter=1,ngpu=1):
        """
        Function to cropping an input layer
        :param outfilter: number of output units (1: for binary output; n for multiclass output);
        :param ngpu: number of GPU, default 1;
        :return: the model constructed with the input and output layer
        """
        inputlayer = Input(shape=self.input_shape,name='img')
        self.inputlayer = inputlayer
        self.nmaxpool=0
        if self.nfeature==0:
            outlayer = self.levelsBlock(inputlayer, self.start_filter, self.depth, self.dropout)
            outlayer = self.cropLayer(outlayer)
            outlayer = Conv2D(outfilter, 1, activation='sigmoid')(outlayer)
            return Model(inputs=inputlayer, outputs=outlayer)
        input_fea_layer = Input((self.nfeature,), name='feat')
        outlayer = self.levelsBlock(inputlayer, self.start_filter, self.depth, self.dropout,input_fea_layer)
        outlayer=self.cropLayer(outlayer)
        outlayer = Conv2D(outfilter, 1, activation='sigmoid',kernel_initializer=self.k_initializer)(outlayer)
        if ngpu>1:
            with tf.device("/cpu:0"):
                fmodel = Model(inputs=[inputlayer, input_fea_layer], outputs=outlayer)
            fmodel = multi_gpu_model(fmodel, gpus=ngpu)
            return fmodel
        return Model(inputs=[inputlayer, input_fea_layer], outputs=outlayer)




import os
import re
import sys
import platform
import subprocess

import setuptools
from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion

from os import path
this_directory = path.abspath(path.dirname(__file__))
import io
with io.open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

__version__ = '0.1.62'

class get_pybind_include(object):
    """Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)

class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " +
                               ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmake_version = LooseVersion(re.search(r'version\s*([\d.]+)', out.decode()).group(1))
            if cmake_version < '3.1.0':
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable]

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        if platform.system() == "Windows":
            cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '-j2']

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp)

class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

    if sys.platform == 'darwin':
        c_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)


from subprocess import CalledProcessError 

ext=Extension(
        'resmcseg._metrics',
        ['src/bindings.cpp','src/cmetrics.cpp'],
        include_dirs=[
            # Path to pybind11 headers
            'src',
            get_pybind_include(),
            get_pybind_include(user=True)
        ],
        language='c++'
    )
# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14] compiler flag.
    The c++14 is prefered over c++11 (when it is available).
    """
    #if has_flag(compiler, '-std=c++14'):
#        return '-std=c++14'
    if has_flag(compiler, '-std=c++1y'):
        return '-std=c++1y'
    elif has_flag(compiler, '-std=c++11'):
        return '-std=c++11'
    else:
        raise RuntimeError('Unsupported compiler -- at least C++11 support '
                           'is needed!')


kwargs = dict(
    name='resmcseg',
    version=__version__,
    author='Lianfa Li',
    author_email='lspatial@gmail.com',
    description='Library for Deep Residual Multiscale Segmenter',
    long_description=long_description,
    long_description_content_type='text/markdown',
    ext_modules=[ext],
    cmdclass=dict(build_ext=BuildExt),
    zip_safe=False, 
    packages=['resmcseg','resmcseg.data','resmcseg.util','resmcseg.model'],
    #packages=['resmcseg'],
    #find_packages(where='.', exclude=(), include=('*',))
    package_dir={'': '.'},
    install_requires = ['keras>=2.2.2','opencv-contrib-python>=3.4.1','tensorflow','numpy','sklearn'],
    include_package_data=True,
    package_data={'resmcseg':['data/*.tif']},
    #data_files=[('tif', ['data/zh17.tif', 'data/zh17_gt.tif'])],
)

# likely there are more exceptions, take a look at yarl example 
try: 
    setup(**kwargs)   
except CalledProcessError: 
    print('Failed to build extension!') 
    del kwargs['ext_modules'] 
    setup(**kwargs) 





from resmcseg.model.gresmcseg import gResMCSeg
from resmcseg.model.gresmcseg_pre import gResMCSegPre
from resmcseg.model.pretrainedmodel import downloadPretrainedModel
from resmcseg.model.resizelayer import ResizeLayer
